import logging
import tempfile
from dataclasses import InitVar, dataclass, field
from datetime import date,datetime
from pathlib import Path
from typing import Iterable, List, Optional, Tuple, Union

import pyproj
import requests
from geopandas import GeoSeries
from pandas import DataFrame
from shapely.geometry import MultiPolygon, Point, Polygon

from wiwb.api_calls import Request
from wiwb.constants import (
    DATA_FORMAT_CODES,
    FILE_SUFFICES,
    INTERVAL_TYPES,
    get_defaults,
)
from wiwb.converters import snake_to_pascal_case
from wiwb.sample import sample_netcdf

logger = logging.getLogger(__name__)
defaults = get_defaults()

@dataclass
class Extent:
    f"""Extent for Settings in request body in correct epsg: {defaults.crs}.

    Parameters
    ----------
    xll : float
        The x-coordinate of the lower-left corner of the extent. Defaults to {defaults.bounds[0]}.
    yll : float
        The y-coordinate of the lower-left corner of the extent. Defaults to {defaults.bounds[1]}.
    xur : float
        The x-coordinate of the upper-right corner of the extent. Defaults to {defaults.bounds[2]}.
    yur : float
        The y-coordinate of the upper-right corner of the extent. Defaults to {defaults.bounds[3]}.
    """

    xll: float = defaults.bounds[0]
    yll: float = defaults.bounds[1]
    xur: float = defaults.bounds[2]
    yur: float = defaults.bounds[3]

    def __post_init__(self):
        if self.width <= 0:
            raise ValueError(
                f"'xll' ({self.xll}) should be smaller than 'xur' ({self.xur})"
            )

        if self.height <= 0:
            raise ValueError(
                f"'yll' ({self.yll}) should be smaller than 'yur' ({self.yur})"
            )

        self.correct_bounds()

    @property
    def width(self):
        return self.xur - self.xll

    @property
    def height(self):
        return self.yur - self.yll

    @property
    def crs(self):
        return pyproj.CRS(self.epsg)

    @property
    def epsg(self):
        return defaults.crs

    @property
    def spatial_reference(self):
        return {"Epsg": self.epsg}

    def correct_bounds(self):
        # get crs-unit
        units = None
        crs_dict = self.crs.to_dict()
        if "unit" in crs_dict.keys():
            units = crs_dict["units"]

        # get min width and height
        if units == "m":
            min_width_height = 10
        else:
            min_width_height = 0.0001  # we assume degrees

        # alter bounds
        if self.width < min_width_height:
            logger.warning(
                f"""Width of bounds < min_width ({self.width < min_width_height}). {self.xll} and {self.xur} will be adjusted"""  # noqa:E501
            )
            self.xll -= (min_width_height - self.width) / 2
            self.xur += (min_width_height - self.width) / 2

        if self.height < min_width_height:
            logger.warning(
                f"""Height of bounds < min_height ({self.height < min_width_height}). {self.yll} and {self.yur} will be adjusted"""  # noqa:E501
            )
            self.yll -= (min_width_height - self.height) / 2
            self.yur += (min_width_height - self.height) / 2

    def json(self):
        dict = self.__dict__.copy()
        dict["spatial_reference"] = self.spatial_reference
        return {snake_to_pascal_case(k): v for k, v in dict.items()}


@dataclass
class Interval:
    """Interval for Settings in request body.

    Parameters
    ----------
    type : str
        The interval, either "Days", "Hours", "Minutes", "None"
    value: int
        Increment of the interval

    Example
    -------

    Interval(type="Hours", value=2)

    Is an interval of 2 hours.

    """

    type: INTERVAL_TYPES
    value: int

    def __post_init__(self):
        self.validate()

    def validate(self):
        if self.type not in INTERVAL_TYPES.__args__:
            raise ValueError(
                f"{self.type} not a valid interval-type: {INTERVAL_TYPES.__args__}"
            )

    def json(self):
        self.validate()
        return {snake_to_pascal_case(k): v for k, v in self.__dict__.items()}


@dataclass
class ReaderSettings:
    """WIWB reader-settings

    Parameters
    ----------
    start_date: datetime.date
        Optional Reader start_date
    end_date: datetime.date
        Optional Reader end_date
    model_run: datetime
        Optional datetime for model_run. Defaults to Last
    variable_codes: List[str]
        List of WIWB variable codes
    model_date: datetime.date
        Reader end_date
    structure_type: str
        Optional StructureType for reader
    location_codes: List[str]
        Optional list of location-codes to retrieve data for
    interval: Interval
        Optional time-interval for reader
    extent: Extend
        Optional extent for reader
    """

    variable_codes: list
    start_date: Union[date, None] = None
    end_date: Union[date, None] = None
    model_run: Union[datetime, None, str] = None
    location_codes: Union[list, None] = None
    model_date: Union[date, None] = None
    interval: Union[Interval, None] = None
    extent: Union[Extent, None] = None
    structure_type: Union[str, None] = None
    
    def json(self):
        dict = self.__dict__.copy()
        for k,v in dict.items():
            if isinstance(v, datetime) or isinstance(v, date):
                dict[k] = v.strftime("%Y%m%d%H%M%S")
            elif hasattr(dict[k], "json"):
                dict[k] = dict[k].json()

        return {snake_to_pascal_case(k): v for k, v in dict.items() if v is not None}


@dataclass
class Reader:
    """WIWB reader

    Parameters
    ----------
    data_source_code: str
        WIWB datasourcecode to read
    settings: Settings
        WIWB reader settings
    """

    data_source_code: str
    settings: Union[ReaderSettings, None] = field(default_factory=ReaderSettings)

    def json(self):
        dict = self.__dict__.copy()
        dict["settings"] = dict["settings"].json()

        return {snake_to_pascal_case(k): v for k, v in dict.items()}


@dataclass
class ExporterSettings:
    """WIWB export settings

    Parameters
    ----------
    export_projection_file: bool, optional
        To write a projection file (in case of ASCII Grid). Default is False
    digits_to_round: int, optional
        To round digits to decimals

    """

    export_projection_file: Union[bool, None] = False
    digits_to_round: Union[int, None] = None


    def json(self):
        return {
            snake_to_pascal_case(k): v
            for k, v in self.__dict__.items()
            if v is not None
        }


@dataclass
class Exporter:
    f"""WIWB exporter

    Parameters
    ----------
    data_format_code: {DATA_FORMAT_CODES}, optional
        data-format code to export data to. Defaults to geotiff
    settings: ExporterSettings
        WIWB exporter settings

    """

    data_format_code: DATA_FORMAT_CODES = "geotiff"
    settings: Union[ExporterSettings, None] = None

    def __post_init__(self):
        self.validate()

    def validate(self):
        if self.data_format_code not in DATA_FORMAT_CODES.__args__:
            raise ValueError(
                f"{self.data_format_code} not a valid data-format-code: {DATA_FORMAT_CODES.__args__}"
            )

    def json(self):
        self.validate()
        dict = self.__dict__.copy()
        for k,v in dict.items():
            if hasattr(dict[k], "json"):
                dict[k] = dict[k].json()

        return {snake_to_pascal_case(k): v for k, v in dict.items() if v is not None}


@dataclass
class RequestBody:
    """GetGrids request Body"""

    readers: List[Reader]
    exporter: Union[Exporter, None] = None

    def json(self):
        dict = self.__dict__.copy()
        dict["readers"] = [i.json() for i in dict["readers"]]
        if dict["exporter"] is None:
            dict.pop("exporter")
        else:
            dict["exporter"] = dict["exporter"].json()
        return {snake_to_pascal_case(k): v for k, v in dict.items()}


"""Authorization for the WIWB API"""

from dataclasses import dataclass, field
from typing import Union

import jwt
import requests

from wiwb.constants import AUTH_URL, CLIENT_ID, CLIENT_SECRET

try:
    from datetime import datetime, timedelta, UTC
except ImportError: # support Python 3.9
    from datetime import datetime, timezone, timedelta
    UTC = timezone.utc


@dataclass
class Auth:
    f"""Authorization class for WIWB API.

    Will provide valid headers and tokens for the WIWB API

    Attributes
    ----------
    client_id : str
        A valid client_id for the WIWB API. If not provided it will be read from the os environment
         variable `wiwb_client_id`. By default {CLIENT_ID}.
    client_secret : str
        A valid client_secret for the WIWB API. If not provided it will be read from the os environment
         variable `wiwb_client_secret`. By default {CLIENT_SECRET}.
    url: str
        A valid WIWB token url. By default {AUTH_URL}.
    token: str
        A valid WIWB access token

    Examples
    --------
    from wiwb import Auth

    >>> auth = Auth() # initialize wiwb authorization
    >>> auth.token  # returns a valid token
    'eyJhbGciOiJS...............`

    >>> auth.headers # returns valid headers for requests to the WIWB API
    {{"content-type": "application/json", "Authorization": 'eyJhbGciOiJS...............`}}
    """
    client_id: str = CLIENT_ID
    client_secret: str = CLIENT_SECRET
    url: str = AUTH_URL
    _token: Union[str, None] = field(default=None, repr=False)

    def __post_init__(self):

        # check if client_id and client_secret are valid
        if self.client_id is None:
            raise ValueError(
                f"Invalid 'client_id': '{self.client_id}'. "
                f"Provide at init or specify 'wiwb_client_id' as os environment variable"  # noqa:E501
            )

        if self.client_secret is None:
            raise ValueError(
                f"Invalid 'client_secret':  '{self.client_secret}'."
                f"Provide at init or specify 'wiwb_client_id' as os environment variable"  # noqa:E501
            )

        # get a token to get started
        self._get_token()

    @property
    def token(self) -> str:
        """Return a valid WIWB access token."""
        if not self.is_token_valid:
            self._get_token()
        return self._token

    @property
    def is_token_valid(self) -> bool:
        """Check if current token is still valid."""
        token_decoded = jwt.decode(self._token, options={"verify_signature": False})
        token_exp_datetime = datetime.fromtimestamp(token_decoded["exp"], UTC)
        # token_exp_datetime = datetime.utcfromtimestamp(token_decoded["exp"])
        current_datetime = datetime.now(UTC) - timedelta(minutes=1)
        # current_datetime = datetime.utcnow() - timedelta(minutes=1)
        return current_datetime < token_exp_datetime

    @property
    def headers(self) -> dict:
        """Headers for WIWB API requests"""
        return {
            "content-type": "application/json",
            "Authorization": "Bearer " + self.token,
        }

    def _get_token(self) -> None:
        """Get, and store, a fresh WIWB access token"""
        response = requests.post(
            self.url,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "grant_type": "client_credentials",
            },
        )
        if response.ok:
            self._token = response.json()["access_token"]
        else:
            response.raise_for_status()



from datetime import date
from pathlib import Path
from typing import List, Union

import numpy as np
import pandas as pd
import xarray
from affine import Affine
from geopandas import GeoSeries
from numpy import ndarray
from rasterstats import zonal_stats
from typing import Union


def flatten_stats(stats_dict: List[str], stats: List[str]) -> List[float]:
    return np.array([[item[stat] for stat in stats] for item in stats_dict]).flatten()


def sample_geoseries(
    values: ndarray,
    geometries: Union[List, GeoSeries],
    affine: Affine,
    nodata: float,
    stats: Union[str, List[str]] = "mean",
) -> List[float]:

    # we work with lists to flatten_stats
    if isinstance(stats, str):
        stats = [stats]

    stats_dict = zonal_stats(
        geometries,
        values,
        affine=affine,
        nodata=nodata,
        stats=" ".join(stats),
        boundless=True,
    )

    return flatten_stats(stats_dict, stats)


def sample_netcdf(
    nc_file: Union[Path, str],
    variable_code: str,
    geometries: Union[List, GeoSeries],
    stats: Union[str, List[str]] = "mean",
    start_date: Union[date, None] = None,
    end_date: Union[date, None] = None,
    unlink: bool = False,
) -> pd.DataFrame:
    """Sample a set of geometries over a netcdf file

    Parameters
    ----------
    nc_file : Path or str
        path to NetCDF file
    variable_code : str
        Variable in NetCDF file to sample
    geometries : Union[List, GeoSeries]
        geometries to sample
    stats : List[str]
        statistics to sample
    start_date : Union[date, None]
        start date for selection, by default None
    end_date: Union[date, None]
        end date for selection, by default None
    unlink : bool, optional
        option to delete netcdf-file after sampling, by default False

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame with statistics per timestamp per geometry
    """
    if isinstance(nc_file, str):
        nc_file = Path(str)
    assert nc_file.is_file(), f"nc_file {nc_file} does not exist"

    # read temp-source for sampling
    with xarray.open_dataset(nc_file, engine="netcdf4") as ds:
        if (start_date is not None) and (end_date is not None):
            ds = ds.sel(time=slice(start_date, end_date))
        nodata =  ds[variable_code].encoding.get("_FillValue", None)
        affine = ds.rio.transform()
        data = {
            time: sample_geoseries(
                values=ds[variable_code].sel(time=time).values,
                geometries=geometries,
                affine=affine,
                nodata=nodata,
                stats=stats,
            )
            for time in ds["time"].values
        }

    # delete temp-file
    if unlink:
        if nc_file.exists():
            nc_file.unlink()

    # create columns
    if isinstance(stats, str):
        stats = [stats]

    if len(stats) == 1:
        columns = geometries.index
    else:
        geom_index = geometries.index
        columns = pd.MultiIndex.from_product(
            iterables=[geom_index, stats], names=["index", "stats"]
        )

    return pd.DataFrame.from_dict(data, orient="index", columns=columns)


def sample_netcdfs(
    nc_files: list[Path],
    variable_code: str,
    geometries: Union[List, GeoSeries],
    stats: Union[str, List[str]] = "mean",
    start_date: Union[date, None] = None,
    end_date: Union[date, None] = None,
):
    """Sample over a set of netcdf-files

    Parameters
    ----------
    nc_files : list[Path]
        A list of netcdf-files
    variable_code : str
        Variable in NetCDF file to sample
    geometries : Union[List, GeoSeries]
        geometries to sample
    stats : List[str]
        statistics to sample
    start_date : Union[date, None]
        start date for selection, by default None
    end_date: Union[date, None]
        end date for selection, by default None

    Returns
    ------
    pd.DataFrame
        Pandas DataFrame with statistics per timestamp per geometry
    """
    assert nc_files, f"no NetCDF files to sample"
    for nc_file in nc_files:
        assert nc_file.is_file(), f"nc_file {nc_file} does not exist"

    # read all transforms to see if dataset is consistent
    transforms = list(
        set(xarray.open_dataset(file).rio.transform() for file in nc_files)
    )
    if len(transforms) == 1:
        with xarray.open_dataset(nc_files[0], decode_coords="all") as ds:
            geometries = geometries.to_crs(ds.rio.crs)
    else:
        raise ValueError(
            f"Files do not have one consistent transform. Got {transforms}"
        )

    dfs = [
        sample_netcdf(
            nc_file,
            variable_code,
            geometries,
            stats=stats,
            start_date=start_date,
            end_date=end_date,
            unlink=False,
        )
        for nc_file in nc_files
    ]
    dfs = [i for i in dfs if not i.empty]
    df = pd.concat(dfs).sort_index()

    return df


def sample_nc_dir(
    dir_path: Union[Path, str],
    variable_code: str,
    geometries: Union[List, GeoSeries],
    stats: Union[str, List[str]] = "mean",
    start_date: Union[date, None] = None,
    end_date: Union[date, None] = None,
):
    """Sample over a directory of netcdf-files

    Parameters
    ----------
    dir_path : Union[Path, str]
        Directory with netcdf files
    variable_code : str
        Variable in NetCDF file to sample
    geometries : Union[List, GeoSeries]
        geometries to sample
    stats : List[str]
        statistics to sample
    start_date : Union[date, None]
        start date for selection, by default None
    end_date: Union[date, None]
        end date for selection, by default None

    Returns
    ------
    pd.DataFrame
        Pandas DataFrame with statistics per timestamp per geometry
    """
    if isinstance(dir_path, str):
        dir_path = Path(dir_path)
    assert dir_path.is_dir(), f"dir_path {dir_path} does not exist"

    nc_files = [x for x in dir_path.iterdir() if x.suffix == ".nc"]

    df = sample_netcdfs(
        nc_files,
        variable_code=variable_code,
        geometries=geometries,
        stats=stats,
        start_date=start_date,
        end_date=end_date,
    )
    return df



from datetime import date
from pathlib import Path

from wiwb.sample import sample_nc_dir

START_DATE = date(2015, 1, 1)
END_DATE = date(2015, 1, 2)
DIR = Path(__file__).parent.joinpath("data")
STATS = ["mean", "min", "max"]


def test_sample_nc_dir(geoseries, nc_df):
    variables = [i.name for i in DIR.glob(r"*/")]
    variable = variables[0]
    dir = DIR.joinpath(variable)

    df = sample_nc_dir(dir, variable, geoseries, STATS, START_DATE, END_DATE)

    assert not df.empty
    assert (df * 100).astype(int).equals(nc_df)



import logging
import tempfile
from dataclasses import InitVar, dataclass, field
from datetime import date
from pathlib import Path
from typing import Iterable, List, Optional, Tuple, Union

import pyproj
import requests
from geopandas import GeoSeries
from pandas import DataFrame
from shapely.geometry import MultiPolygon, Point, Polygon

from wiwb.api_calls import Request
from wiwb.auth import Auth
from wiwb.api_calls.body import RequestBody, ReaderSettings, Interval, Extent, Exporter, Reader
from wiwb.constants import (
    DATA_FORMAT_CODES,
    FILE_SUFFICES,
    INTERVAL_TYPES,
    get_defaults,
)
from wiwb.converters import rename_file
from wiwb.sample import sample_netcdf
import zipfile

logger = logging.getLogger(__name__)
defaults = get_defaults()

# @dataclass
class GetGrids(Request):
    """GetGrids request"""

    def __init__(
            self,
            auth: Auth,
            base_url: str,
            data_source_code: str,
            variable_code:str,
            start_date: date,
            end_date:date,
            unzip: bool=True,
            interval:Tuple[str, int] = ("Hours", 1),
            data_format_code: DATA_FORMAT_CODES = "geotiff",
            geometries: Union[GeoSeries, Iterable[Union[Point, Polygon, MultiPolygon]] | None] = None,
            bounds: Union[Tuple[float, float, float, float] |  None] = None
            ):

        # init from Requests class
        super().__init__(auth=auth, base_url=base_url)
    
        # GetGrids fields
        self.data_source_code: str = data_source_code
        self.variable_code:str = variable_code
        self.start_date: date = start_date
        self.end_date:date = end_date
        self.unzip: bool = unzip
        self.interval:Tuple[str, int] = interval
        self.data_format_code: DATA_FORMAT_CODES = data_format_code

        # hidden variables
        self._response: Union[requests.Response, None] = None
        self._geoseries: Union[GeoSeries, None] = None
        self._bounds: Union[Tuple[float, float, float, float], None] = field(
            init=False, default=None
        )

        # Set geometries and calculate bounds
        self.set_geometries(geometries)
        self.set_bounds(bounds)

    @property
    def epsg(self):
        return defaults.crs

    @property
    def crs(self):
        return self.body.readers[0].settings.extent.crs

    @property
    def body(self) -> RequestBody:
        reader_settings = ReaderSettings(
            start_date=self.start_date,
            end_date=self.end_date,
            variable_codes=[self.variable_code],
            interval=Interval(*self.interval),
            extent=Extent(*self.bounds),
        )

        reader = Reader(self.data_source_code, settings=reader_settings)

        exporter = Exporter(data_format_code=self.data_format_code)

        return RequestBody(readers=[reader], exporter=exporter)

    @property
    def bbox(self):  # noqa:F811
        return self._bounds

    @property
    def file_name(self):
        stem = "_".join(
            [
                self.data_source_code,
                self.variable_code,
                self.start_date.strftime('%Y%m%dT%H%M%S'),
                self.end_date.strftime('%Y%m%dT%H%M%S'),
            ]
        )
        suffix = FILE_SUFFICES[self.data_format_code]
        return f"{stem}.{suffix}"

    @property
    def geoseries(self) -> GeoSeries:
        return self._geoseries

    @property
    def url_post_fix(self) -> str:
        return "grids/get"
    
    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        return self._get_bounds(self._bounds)

    def _to_geoseries(
        self,
        geometries: Optional[Union[GeoSeries, Iterable[Union[Point, Polygon, MultiPolygon]]]],
    ) -> GeoSeries:

        # convert iterable to GeoSeries
        if geometries is not None:
            if not isinstance(geometries, GeoSeries):
                geometries = GeoSeries(geometries)

            # Check if geometries are Point, Polygon, or MultiPolygon
            if not all(
                (
                    i in ["Point", "Polygon", "MultiPolygon"]
                    for i in geometries.geom_type
                )
            ):
                raise ValueError(
                    f"Geometries must be Point, Polygon, or MultiPolygon, got {geometries.geom_type.unique()}"
                )

        geometries = self._reproject_geoseries(geoseries=geometries)
        return geometries

    def _reproject_geoseries(self, geoseries: GeoSeries) -> GeoSeries:
        """Set or reproject geoseries to self.epsg"""
        if geoseries.crs is None:
            logger.warning(f"no crs specified in geoseries, will be set to {self.epsg}")
            geoseries.crs = self.epsg
        else:
            geoseries = geoseries.to_crs(self.epsg)
        return geoseries

    def _get_bounds(self, bounds: Union[Tuple[float, float, float, float], None]):
        if (
            self._geoseries is not None
        ):  # if geometries are specified, we'll get bounds from geometries
            bounds = tuple(self._geoseries.total_bounds)
            if bounds is None:
                logger.warning(
                    "bounds will be ignored as long as geometries are not None"
                )
        elif bounds is None:  # if geometries aren't specified, user has to set bounds
            raise ValueError(
                """Specify either 'geometries' or 'bounds', both are None"""
            )
        return bounds

    def run(self):
        self._response = None
        self._response = requests.post(
            self.url, headers=self.auth.headers, json=self.body.json()
        )

        if not self._response.ok:
            self._response.raise_for_status()

    def set_geometries(
        self,
        geometries: Optional[Union[GeoSeries, Iterable[Union[Point, Polygon, MultiPolygon]]]],
    ) -> None:
        """Set a list or GeoSeries with Point, Polygon or MultiPolygon values. Handles conversion to
        GeoSeries and reprojection

        Parameters
        ----------
        geometries : GeoSeries, Iterable[Union[Point, Polygon, MultiPolygon]]
            A list or GeoSeries with Point, Polygon and Multipolygon objects
        """
        if geometries is not None:
            geoseries = self._to_geoseries(geometries)
            self._geoseries = geoseries
        else:
            self._geoseries = geometries

    def set_bounds(self, bounds: Tuple[float, float, float, float] | None) -> None:
        """Set new bounds values. Fits bounds to geoseries.bounds

        Parameters
        ----------
        bounds : Tuple[float, float, float, float]
            Bounds tuple

        """

        bounds = self._get_bounds(bounds)
        self._bounds = bounds

    def write_tempfile(self):
        with tempfile.NamedTemporaryFile(
            suffix=FILE_SUFFICES[self.data_format_code], delete=False
        ) as tmp_file:
            tmp_file_path = Path(tmp_file.name)
            tmp_file.write(self._response.content)
        return tmp_file_path

    def sample(self, stats: Union[str, List[str]] = "mean") -> DataFrame:
        """Sample statistics per geometry

        Parameters
        ----------
        stats : Union[str, List[str]]
            statistics to sample, provided as list of statistics or a string with one statistic. defaults to mean

            All stats in rasterstats.zonal_stats are available: https://pythonhosted.org/rasterstats/manual.html#statistics
            Common values are:
                - mean: average value of all cells in polygon
                - max: maximum value of all cells in polygon
                - min: minimum value of all cells in polygon
                - percentile_#: percentile value of all cells in polygon. E.g. percentile_50, gives 50th percentile (median) value

            Notes:
            - Providing multiple values, will create a multi-index column in your dataframe
            - Providing multiple statistics, as specified above, doesn't make much sense as it will always return the same value
        """  # noqa:E501

        # check if geometries are set
        if self._geoseries is None:
            raise TypeError(
                """'geometries' is None, should be list or GeoSeries. Set it first"""
            )

        # check if data_format_code is netcdf
        if self.data_format_code != "netcdf4.cf1p6":
            self.data_format_code = "netcdf4.cf1p6"
            self.run()

        # re-run
        if self._response is None:
            self.run()

        # write content in temp-file
        temp_file = self.write_tempfile()

        # sample temp_file
        df = sample_netcdf(
            nc_file=temp_file,
            variable_code=self.variable_code,
            geometries=self.geoseries,
            stats=stats,
            unlink=True,
        )

        return df

    def to_directory(self, output_dir: Union[str, Path], unzip:bool=False, rename_map:dict| None = None):
        """Write response.content to an output-file"""
        if self._response is None:
            self.run()

        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        if unzip:

            # write content in temp_file
            temp_file = self.write_tempfile()

            with zipfile.ZipFile(temp_file, "r") as zf:
                for info in zf.infolist():
                    if info.is_dir():
                        continue

                    file_name = info.filename
                    if rename_map is not None:
                        file_name= output_dir / rename_file(file_name, **rename_map)
                    target = output_dir / file_name

                    # Schrijf binaire inhoud veilig weg
                    with zf.open(info, "r") as src, open(target, "wb") as dst:
                        while True:
                            chunk = src.read(1024 * 1024)
                            if not chunk:
                                break
                            dst.write(chunk)

        else:
            output_file = output_dir / self.file_name
            output_file.write_bytes(self._response.content)



from wiwb.api_calls.base import Request
from wiwb.api_calls.get_data_sources import GetDataSources
from wiwb.api_calls.get_grids import GetGrids
from wiwb.api_calls.get_variables import GetVariables
from wiwb.api_calls.body import RequestBody, ReaderSettings, Interval, Extent, Exporter, Reader, ExporterSettings

__all__ = ["Request", "GetDataSources", "GetVariables", "GetGrids", "RequestBody", "ReaderSettings", "Interval", "Extent", "Exporter", "Reader", "ExporterSettings"]



# %%
import os
from typing import Literal

from geopandas import GeoSeries
from shapely.geometry import MultiPolygon, Point, Polygon, box
import sys

API_URL = "https://wiwb.hydronet.com/api"
AUTH_URL = (
    "https://login.hydronet.com/auth/realms/hydronet/protocol/openid-connect/token"
)

CLIENT_ID = os.getenv("wiwb_client_id")
CLIENT_SECRET = os.getenv("wiwb_client_secret")

PYTHON_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"

FILE_SUFFICES = {
    "geotiff": "zip",
    "aaigrid": "hdf5",
    "hdf5": "hdf5",
    "netcdf4.cf1p6": "nc",
    "netcdf4.cf1p6.zip": "zip",
}

PRIMARY_STRUCTURE_TYPES = Literal[
    "EnsembleGrid",
    "EnsembleTimeSeries",
    "Event",
    "Grid",
    "ModelGrid",
    "ModelTimeSeries",
    "TimeSeries",
]

DATA_FORMAT_CODES = Literal["geotiff", "aaigrid", "hdf5", "netcdf4.cf1p6", "netcdf4.cf1p6.zip", "hydronet.csv.simple", "hydronet.csv.simple", "json"]

IMPLEMENTED_GEOMETRY_TYPES = [Point, Polygon, MultiPolygon]

INTERVAL_TYPES = Literal["Days", "Hours", "Minutes", "None"]

CRS_EPSG = 28992
LL_POINT = Point(119865, 449665)
UR_POINT = Point(127325, 453565)
OTHER_POINT = Point(135125, 453394)
POLYGON = box(LL_POINT.x, LL_POINT.y, UR_POINT.x, UR_POINT.y)
GEOSERIES = GeoSeries(
    [LL_POINT, UR_POINT, OTHER_POINT, POLYGON],
    index=["ll_point", "ur_point", "other_point", "polygon"],
    crs=CRS_EPSG,
)


class Defaults:
    bounds: tuple[float, float, float, float] = (109950, 438940, 169430, 467600)
    crs: int = 28992
    geoseries = GEOSERIES


def get_defaults(**kwargs):
    return Defaults(**kwargs)


# %%



"""For converting one thing into another"""
from pathlib import Path
from datetime import datetime


def snake_to_pascal_case(snake_case: str) -> str:
    """Convert snake_case to PascalCase."""
    words = snake_case.split("_")
    return "".join(i.title() for i in words)


def rename_file(file_name: Path, prefix:str, use: str = "start", datetime_pattern: str = "%Y%m%d%H%M"):
    """Rename a WIWB output file to a different datetime pattern"""
    file_name = Path(file_name)
    _, _, start_str, end_str = file_name.stem.split("_")
    if use == "start":
        date_time = datetime.strptime(start_str, "%Y-%m-%dT%Hh%Mm%Ss")
    elif use == "end":
        date_time = datetime.strptime(end_str, "%Y-%m-%dT%Hh%Mm%Ss")
    else:
        raise ValueError(f"value for use {use} not implemented. Use start or end")

    return f"{prefix}_{date_time.strftime(datetime_pattern)}{file_name.suffix}"
